@@ -15,7 +15,7 @@
 #include <string>
 
 #include <LOOLWebSocket.hpp>
-#include "net/WebSocketHandler.hpp"
+#include "net/Socket.hpp"
 
 class UnitBase;
 class UnitWSD;
@@ -24,9 +24,13 @@ class UnitTimeout;
 class UnitHTTPServerRequest;
 class UnitHTTPServerResponse;
 
+class WebSocketHandler;
+
 // Forward declaration to avoid pulling the world here.
 namespace Poco
 {
+    class MemoryInputStream;
+
     namespace Net
     {
         class HTTPServerRequest;
@@ -95,6 +99,12 @@ public:
         return false;
     }
 
+    /// Message that is about to be sent via the websocket.
+    virtual bool filterSendMessage(const char* /* data */, const size_t /* len */, const WSOpCode /* code */, const bool /* flush */, int& /*unitReturn*/)
+    {
+        return false;
+    }
+
     /// Hook the disk space check
     virtual bool filterCheckDiskSpace(const std::string & /* path */,
                                       bool & /* newResult */)
@@ -102,6 +112,18 @@ public:
         return false;
     }
 
+    /// Trap and filter alerting all users
+    virtual bool filterAlertAllusers(const std::string & /* msg */)
+    {
+        return false;
+    }
+
+    /// Custom response to a http request.
+    virtual bool handleHttpRequest(const Poco::Net::HTTPRequest& /*request*/, Poco::MemoryInputStream& /*message*/,std::shared_ptr<StreamSocket>& /*socket*/)
+    {
+        return false;
+    }
+
     /// If the test times out this gets invoked, the default just exits.
     virtual void timeout();
 
@@ -202,6 +224,12 @@ public:
         return false;
     }
 
+    /// To force the save operation being handled as auto-save from a unit test.
+    virtual bool isAutosave()
+    {
+        return false;
+    }
+
     // ---------------- WSD events ----------------
     virtual void onChildConnected(const int /* pid */, const std::string& /* sessionId */) {}
     /// When admin notify message is sent
