@@ -3,10 +3,11 @@
  * L.Socket contains methods for the communication with the server
  */
 
-/* global _ vex $ errorMessages */
+/* global _ vex $ errorMessages Uint8Array brandProductName brandProductFAQURL */
 L.Socket = L.Class.extend({
 	ProtocolVersionNumber: '0.1',
 	ReconnectCount: 0,
+	WasShownLimitDialog: false,
 
 	getParameterValue: function (s) {
 		var i = s.indexOf('=');
@@ -31,10 +32,16 @@ L.Socket = L.Class.extend({
 			wopiSrc = '?WOPISrc=' + map.options.wopiSrc + '&compat=/ws';
 		}
 
+		var websocketURI = map.options.server + '/lool/' + encodeURIComponent(map.options.doc + '?' + $.param(map.options.docParams)) + '/ws' + wopiSrc;
 		try {
-			this.socket = new WebSocket(map.options.server + '/lool/' + encodeURIComponent(map.options.doc + '?' + $.param(map.options.docParams)) + '/ws' + wopiSrc);
+			if (this.socket) {
+				this.close();
+			}
+			this.socket = new WebSocket(websocketURI);
 		} catch (e) {
-			this._map.fire('error', {msg: _('Oops, there is a problem connecting to LibreOffice Online : ' + e), cmd: 'socket', kind: 'failed', id: 3});
+			this._map.fire('error', {msg: _('Oops, there is a problem connecting to LibreOffice Online : ').replace('LibreOffice Online', (typeof brandProductName !== 'undefined' ? brandProductName : 'LibreOffice Online')) + _('Cannot create websocket, please restart your browser.'), cmd: 'socket', kind: 'failed', id: 3});
+			console.log('Failed to create websocket: ' + websocketURI);
+			console.log('Due to an exception: ' + e);
 			return;
 		}
 
@@ -83,12 +90,20 @@ L.Socket = L.Class.extend({
 	},
 
 	sendMessage: function (msg, coords) {
-		if ((!msg.startsWith('useractive') && !msg.startsWith('userinactive') && !this._map._active) ||
-		    this._map._fatal) {
-			// Avoid communicating when we're inactive.
+		if (this._map._fatal) {
+			// Avoid communicating when we're in fatal state
 			return;
 		}
 
+		if (!this._map._active) {
+			// Avoid communicating when we're inactive.
+			if (typeof msg !== 'string')
+				return;
+
+			if (!msg.startsWith('useractive') && !msg.startsWith('userinactive'))
+				return;
+		}
+
 		var socketState = this.socket.readyState;
 		if (socketState === 2 || socketState === 3) {
 			this._map.loadDocument();
@@ -215,7 +230,7 @@ L.Socket = L.Class.extend({
 		}
 		else if (textMsg.startsWith('lokitversion ')) {
 			var lokitVersionObj = JSON.parse(textMsg.substring(textMsg.indexOf('{')));
-			var h = lokitVersionObj.BuildId.substring(0, 7);
+			h = lokitVersionObj.BuildId.substring(0, 7);
 			if (parseInt(h,16).toString(16) === h.toLowerCase().replace(/^0+/, '')) {
 				h = '<a target="_blank" href="https://hub.libreoffice.org/git-core/' + h + '">' + h + '</a>';
 			}
@@ -258,6 +273,7 @@ L.Socket = L.Class.extend({
 			textMsg = textMsg.substring('close: '.length);
 			msg = '';
 			var postMsgData = {};
+			var showMsgAndReload = false;
 			// This is due to document owner terminating the session
 			if (textMsg === 'ownertermination') {
 				msg = _('Session terminated by document owner');
@@ -288,7 +304,7 @@ L.Socket = L.Class.extend({
 				var max = 10000;
 				var timeoutMs = Math.floor(Math.random() * (max - min) + min);
 
-				socket = this;
+				var socket = this;
 				map = this._map;
 				vex.timer = setInterval(function() {
 					if (socket.connected()) {
@@ -305,12 +321,24 @@ L.Socket = L.Class.extend({
 					try {
 						map.loadDocument(map);
 					} catch (error) {
+						console.warn('Cannot load document.');
 					}
 				}, timeoutMs);
 			}
 			else if (textMsg.startsWith('documentconflict')) {
 				msg = _('Document has changed in storage. Loading the new document. Your version is available as revision.');
+				showMsgAndReload = true;
+			}
+			else if (textMsg.startsWith('versionrestore:')) {
+				textMsg = textMsg.substring('versionrestore:'.length).trim();
+				if (textMsg === 'prerestore_ack') {
+					msg = _('Restoring older revision. Any unsaved changes will be available in version history');
+					this._map.fire('postMessage', {msgId: 'App_VersionRestore', args: {Status: 'Pre_Restore_Ack'}});
+					showMsgAndReload = true;
+				}
+			}
 
+			if (showMsgAndReload) {
 				if (this._map._docLayer) {
 					this._map._docLayer.removeAllViews();
 				}
@@ -326,13 +354,14 @@ L.Socket = L.Class.extend({
 						// Activate and cancel timer and dialogs.
 						map._activate();
 					} catch (error) {
+						console.warn('Cannot activate map');
 					}
 				}, 3000);
 			}
 
 			// Close any open dialogs first.
 			if (vex.dialogID > 0) {
-				var id = vex.dialogID;
+				id = vex.dialogID;
 				vex.dialogID = -1;
 				vex.close(id);
 			}
@@ -368,7 +397,7 @@ L.Socket = L.Class.extend({
 
 			if (textMsg === 'idle' || textMsg === 'oom') {
 				var map = this._map;
-				options.$vex.bind('click.vex', function(e) {
+				options.$vex.bind('click.vex', function() {
 					console.debug('idleness: reactivating');
 					map._documentIdle = false;
 					return map._activate();
@@ -415,24 +444,46 @@ L.Socket = L.Class.extend({
 				// TODO: We really really need to factor this out duplicate dialog code logic everywhere
 				// Close any open dialogs first.
 				if (vex.dialogID > 0) {
-					var id = vex.dialogID;
+					id = vex.dialogID;
 					vex.dialogID = -1;
 					vex.close(id);
 				}
 
-				vex.dialog.confirm({
-					message: _('Document has been changed in storage. Do you want to discard your changes, and load the document from the storage?'),
+				vex.dialog.open({
+					message: _('Document has been changed in storage. What would you like to do with your unsaved changes?'),
+					escapeButtonCloses: false,
+					overlayClosesOnClick: false,
+					contentCSS: { width: '600px' },
 					buttons: [
-						$.extend({}, vex.dialog.buttons.YES, { text: _('Discard changes') }),
-						$.extend({}, vex.dialog.buttons.NO, { text: _('Overwrite document') })
+						$.extend({}, vex.dialog.buttons.YES, { text: _('Discard'),
+						                                      click: function($vexContent) {
+							                                      $vexContent.data().vex.value = 'discard';
+							                                      vex.close($vexContent.data().vex.id);
+						                                      }}),
+						$.extend({}, vex.dialog.buttons.YES, { text: _('Overwrite'),
+						                                      click: function($vexContent) {
+							                                      $vexContent.data().vex.value = 'overwrite';
+							                                      vex.close($vexContent.data().vex.id);
+						                                      }}),
+						$.extend({}, vex.dialog.buttons.YES, { text: _('Save to new file'),
+						                                      click: function($vexContent) {
+							                                      $vexContent.data().vex.value = 'saveas';
+							                                      vex.close($vexContent.data().vex.id);
+						                                      }})
 					],
 					callback: L.bind(function(value) {
-						if (value) {
+						if (value === 'discard') {
 							// They want to refresh the page and load document again for all
 							this.sendMessage('closedocument');
-						} else {
+						} else if (value === 'overwrite') {
 							// They want to overwrite
 							this.sendMessage('savetostorage force=1');
+						} else if (value === 'saveas') {
+							var filename = this._map['wopi'].BaseFileName;
+							if (filename) {
+								filename = L.LOUtil.generateNewFileName(filename, '_new');
+								this._map.saveAs(filename);
+							}
 						}
 					}, this)
 				});
@@ -509,6 +560,7 @@ L.Socket = L.Class.extend({
 						// Activate and cancel timer and dialogs.
 						map._activate();
 					} catch (error) {
+						console.warn('Cannot activate map');
 					}
 				}, 1000);
 			}
@@ -535,31 +587,37 @@ L.Socket = L.Class.extend({
 		}
 		else if (textMsg.startsWith('error:') && !this._map._docLayer) {
 			textMsg = textMsg.substring(6);
-			if (command.errorKind === 'limitreached') {
-				this._map._fatal = true;
-				this._map._active = false; // Practically disconnected.
+			if (command.errorKind === 'hardlimitreached') {
 
-				// Servers configured for 50 documents are not demo/development.
-				if (parseInt(command.params[0]) >= 50) {
-					textMsg = errorMessages.limitreachedprod;
-					textMsg = textMsg.replace(/%0/g, command.params[0]);
-					textMsg = textMsg.replace(/%1/g, command.params[1]);
-				}
-				else {
-					textMsg = errorMessages.limitreached;
-					textMsg = textMsg.replace(/%0/g, command.params[0]);
-					textMsg = textMsg.replace(/%1/g, command.params[1]);
-					textMsg = textMsg.replace(/%2/g, (typeof brandProductName !== 'undefined' ? brandProductName : 'LibreOffice Online'));
-					textMsg = textMsg.replace(/%3/g, (typeof brandProductFAQURL !== 'undefined' ? brandProductFAQURL : 'https://wiki.documentfoundation.org/Development/LibreOffice_Online'));
-				}
+				textMsg = errorMessages.limitreachedprod;
+				textMsg = textMsg.replace(/%0/g, command.params[0]);
+				textMsg = textMsg.replace(/%1/g, command.params[1]);
 			}
 			else if (command.errorKind === 'serviceunavailable') {
-				this._map._fatal = true;
-				this._map._active = false; // Practically disconnected.
 				textMsg = errorMessages.serviceunavailable;
 			}
+			this._map._fatal = true;
+			this._map._active = false; // Practically disconnected.
 			this._map.fire('error', {msg: textMsg});
 		}
+		else if (textMsg.startsWith('info:') && command.errorCmd === 'socket') {
+			if (command.errorKind === 'limitreached' && !this.WasShownLimitDialog) {
+				this.WasShownLimitDialog = true;
+				textMsg = errorMessages.limitreached;
+				textMsg = textMsg.replace(/{docs}/g, command.params[0]);
+				textMsg = textMsg.replace(/{connections}/g, command.params[1]);
+				textMsg = textMsg.replace(/{productname}/g, (typeof brandProductName !== 'undefined' ?
+						brandProductName : 'LibreOffice Online'));
+				var brandFAQURL = (typeof brandProductFAQURL !== 'undefined') ?
+						brandProductFAQURL : 'https://hub.libreoffice.org/professional-online-support';
+				this._map.fire('infobar',
+					{
+						msg: textMsg,
+						action: brandFAQURL,
+						actionLabel: errorMessages.infoandsupport
+					});
+			}
+		}
 		else if (textMsg.startsWith('pong ') && this._map._docLayer && this._map._docLayer._debug) {
 			var times = this._map._docLayer._debugTimePING;
 			var timeText = this._map._docLayer._debugSetTimes(times, +new Date() - this._map._docLayer._debugPINGQueue.shift());
@@ -602,15 +660,16 @@ L.Socket = L.Class.extend({
 				this.ReconnectCount = 0;
 				clearTimeout(vex.timer);
 				if (vex.dialogID > 0) {
-					var id = vex.dialogID;
+					id = vex.dialogID;
 					vex.dialogID = -1;
 					vex.close(id);
 				}
 			}
 		}
-		else if (!textMsg.startsWith('tile:') && !textMsg.startsWith('renderfont:')) {
+		else if (!textMsg.startsWith('tile:') && !textMsg.startsWith('renderfont:') && !textMsg.startsWith('windowpaint:')) {
 			// log the tile msg separately as we need the tile coordinates
 			L.Log.log(textMsg, L.INCOMING);
+
 			if (imgBytes !== undefined) {
 				try {
 					// if it's not a tile, parse the whole message
@@ -628,12 +687,21 @@ L.Socket = L.Class.extend({
 		}
 		else {
 			var data = imgBytes.subarray(index + 1);
-			// read the tile data
-			var strBytes = '';
-			for (var i = 0; i < data.length; i++) {
-				strBytes += String.fromCharCode(data[i]);
+
+			if (data.length > 0 && data[0] == 68 /* D */)
+			{
+				console.log('Socket: got a delta !');
+				var img = data;
+			}
+			else
+			{
+				// read the tile data
+				var strBytes = '';
+				for (var i = 0; i < data.length; i++) {
+					strBytes += String.fromCharCode(data[i]);
+				}
+				img = 'data:image/png;base64,' + window.btoa(strBytes);
 			}
-			var img = 'data:image/png;base64,' + window.btoa(strBytes);
 		}
 
 		if (textMsg.startsWith('status:')) {
@@ -723,7 +791,7 @@ L.Socket = L.Class.extend({
 		// Let onclose (_onSocketClose) report errors.
 	},
 
-	_onSocketClose: function (e) {
+	_onSocketClose: function () {
 		console.debug('_onSocketClose:');
 		var isActive = this._map._active;
 		this._map.hideBusy();
@@ -838,6 +906,25 @@ L.Socket = L.Class.extend({
 			else if (tokens[i].startsWith('wid=')) {
 				command.wireId = this.getParameterValue(tokens[i]);
 			}
+			else if (tokens[i].substring(0, 6) === 'title=') {
+				command.title = tokens[i].substring(6);
+			}
+			else if (tokens[i].substring(0, 12) === 'dialogwidth=') {
+				command.dialogwidth = tokens[i].substring(12);
+			}
+			else if (tokens[i].substring(0, 13) === 'dialogheight=') {
+				command.dialogheight = tokens[i].substring(13);
+			}
+			else if (tokens[i].substring(0, 10) === 'rectangle=') {
+				command.rectangle = tokens[i].substring(10);
+			}
+			else if (tokens[i].substring(0, 12) === 'hiddenparts=') {
+				var hiddenparts = tokens[i].substring(12).split(',');
+				command.hiddenparts = [];
+				hiddenparts.forEach(function (item) {
+					command.hiddenparts.push(parseInt(item));
+				});
+			}
 		}
 		if (command.tileWidth && command.tileHeight && this._map._docLayer) {
 			var defaultZoom = this._map.options.zoom;
