@@ -70,7 +70,7 @@ L.ImpressTileLayer = L.TileLayer.extend({
 		this._draft = null;
 	},
 
-	onAnnotationCancel: function (e) {
+	onAnnotationCancel: function () {
 		if (this._draft) {
 			this._map.removeLayer(this._draft);
 			this._draft = null;
@@ -137,7 +137,7 @@ L.ImpressTileLayer = L.TileLayer.extend({
 		this._map.focus();
 	},
 
-	_onAnnotationZoom: function (e) {
+	_onAnnotationZoom: function () {
 		this.onAnnotationCancel();
 	},
 
@@ -157,12 +157,12 @@ L.ImpressTileLayer = L.TileLayer.extend({
 		this._map.focus();
 	},
 
-	onAnnotationScrollDown: function (e) {
+	onAnnotationScrollDown: function () {
 		this._topAnnotation[this._selectedPart] = Math.min(++this._topAnnotation[this._selectedPart], this._annotations[this._partHashes[this._selectedPart]].length - 1);
 		this.onAnnotationCancel();
 	},
 
-	onAnnotationScrollUp: function (e) {
+	onAnnotationScrollUp: function () {
 		if (this._topAnnotation[this._selectedPart] === 0) {
 			this._map.fire('scrollby', {x: 0, y: -100});
 		}
@@ -170,7 +170,7 @@ L.ImpressTileLayer = L.TileLayer.extend({
 		this.onAnnotationCancel();
 	},
 
-	onUpdateParts: function (e) {
+	onUpdateParts: function () {
 		if (typeof this._prevSelectedPart === 'number') {
 			this.hideAnnotations(this._prevSelectedPart);
 			if (this.hasAnnotations(this._selectedPart)) {
@@ -224,9 +224,7 @@ L.ImpressTileLayer = L.TileLayer.extend({
 	layoutAnnotations: function () {
 		var topAnnotation;
 		var annotations = this._annotations[this._partHashes[this._selectedPart]];
-		var scale = this._map.getZoomScale(this._map.getZoom(), 10);
-		var topRight = this._map.latLngToLayerPoint(this._map.options.maxBounds.getNorthEast())
-			.subtract(this.extraSize.multiplyBy(scale))
+		var topRight = this._map.latLngToLayerPoint(this._map.options.docBounds.getNorthEast())
 			.add(L.point((this._selectedAnnotation ? 3 : 2) * this.options.marginX, this.options.marginY));
 		var bounds, annotation;
 		for (var index in annotations) {
@@ -256,13 +254,13 @@ L.ImpressTileLayer = L.TileLayer.extend({
 						bounds.extend(L.point(bounds.max.x, bounds.max.y + 2 * this.options.marginY));
 					}
 					var offsetX = L.point(2 * this.options.marginX, 0);
-					var topLeft = (bounds ? bounds.getBottomLeft() : topRight).subtract(offsetX);
+					topLeft = (bounds ? bounds.getBottomLeft() : topRight).subtract(offsetX);
 					annotation.setLatLng(this._map.layerPointToLatLng(topLeft));
 					bounds = annotation.getBounds();
 					bounds = L.bounds(bounds.getBottomLeft().add(offsetX), bounds.getTopRight().add(offsetX));
 					bounds.extend(L.point(bounds.max.x, bounds.max.y + 3 * this.options.marginY));
 				} else {
-					var topLeft = bounds ? bounds.getBottomLeft() : topRight;
+					topLeft = bounds ? bounds.getBottomLeft() : topRight;
 					annotation.setLatLng(this._map.layerPointToLatLng(topLeft));
 					annotation.show();
 					bounds = annotation.getBounds();
@@ -298,7 +296,7 @@ L.ImpressTileLayer = L.TileLayer.extend({
 		if (values.comments) {
 			this.clearAnnotations();
 			for (var index in values.comments) {
-				comment = values.comments[index];
+				var comment = values.comments[index];
 				if (!this._annotations[comment.parthash]) {
 					this._annotations[comment.parthash] = [];
 				}
@@ -370,7 +368,7 @@ L.ImpressTileLayer = L.TileLayer.extend({
 
 		var tilePositionsX = '';
 		var tilePositionsY = '';
-		var oldHashes = '';
+		var oldWireIds = '';
 		var needsNewTiles = false;
 
 		for (var key in this._tiles) {
@@ -394,14 +392,14 @@ L.ImpressTileLayer = L.TileLayer.extend({
 						tilePositionsY += ',';
 					}
 					tilePositionsY += tileTopLeft.y;
-					if (oldHashes !== '') {
-						oldHashes += ',';
+					if (oldWireIds !== '') {
+						oldWireIds += ',';
 					}
-					if (this._tiles[key].oldhash === undefined) {
-						oldHashes += '0';
+					if (this._tiles[key].oldWireId === undefined) {
+						oldWireIds += '0';
 					}
 					else {
-						oldHashes += this._tiles[key].oldhash;
+						oldWireIds += this._tiles[key].oldWireId;
 					}
 					needsNewTiles = true;
 					if (this._debug) {
@@ -426,7 +424,7 @@ L.ImpressTileLayer = L.TileLayer.extend({
 				'tileposy=' + tilePositionsY + ' ' +
 				'tilewidth=' + this._tileWidthTwips + ' ' +
 				'tileheight=' + this._tileHeightTwips + ' ' +
-				'oldhash=' + oldHashes;
+				'oldwid=' + oldWireIds;
 
 			this._map._socket.sendMessage(message, '');
 			if (this._debug) {
