@@ -2,7 +2,7 @@
 /*
  * L.TileLayer is used for standard xyz-numbered tile layers.
  */
-
+/* global $ _ Uint8ClampedArray Uint8Array */
 // Implement String::startsWith which is non-portable (Firefox only, it seems)
 // See http://stackoverflow.com/questions/646628/how-to-check-if-a-string-startswith-another-string#4579228
 
@@ -13,17 +13,19 @@ if (typeof String.prototype.startsWith !== 'function') {
 	};
 }
 
+function hex2string(inData)
+{
+	var hexified = [];
+	var data = new Uint8Array(inData);
+	for (var i = 0; i < data.length; i++) {
+		var hex = data[i].toString(16);
+		var paddedHex = ('00' + hex).slice(-2);
+		hexified.push(paddedHex);
+	}
+	return hexified.join('');
+}
+
 L.Compatibility = {
-	clipboardGet: function (event) {
-		var text = null;
-		if (event.clipboardData) { // Standard
-			text = event.clipboardData.getData('text/plain');
-		}
-		else if (window.clipboardData) { // IE 11
-			text = window.clipboardData.getData('Text');
-		}
-		return text;
-	},
 	clipboardSet: function (event, text) {
 		if (event.clipboardData) { // Standard
 			event.clipboardData.setData('text/plain', text);
@@ -87,6 +89,8 @@ L.TileLayer = L.GridLayer.extend({
 		this._visibleCursor = new L.LatLngBounds(new L.LatLng(0, 0), new L.LatLng(0, 0));
 		// Cursor overlay is visible or hidden (for blinking).
 		this._isCursorOverlayVisible = false;
+		// Cursor overlay visibility flag to store last state during zooming
+		this._oldCursorOverlayVisibility = false;
 		// Cursor is visible or hidden (e.g. for graphic selection).
 		this._isCursorVisible = true;
 		// Original rectangle graphic selection in twips
@@ -133,6 +137,14 @@ L.TileLayer = L.GridLayer.extend({
 			});
 		}, this));
 
+		this._dropDownButton = L.marker(new L.LatLng(0, 0), {
+			icon: L.divIcon({
+				className: 'spreadsheet-drop-down-marker',
+				iconSize: null
+			}),
+			interactive: false
+		});
+
 		this._emptyTilesCount = 0;
 		this._msgQueue = [];
 		this._toolbarCommandValues = {};
@@ -245,7 +257,8 @@ L.TileLayer = L.GridLayer.extend({
 		map.on('dragover', this._onDragOver, this);
 		map.on('drop', this._onDrop, this);
 
-		map.on('zoomend', this._onUpdateCursor, this);
+		map.on('zoomstart', this._onZoomStart, this);
+		map.on('zoomend', this._onZoomEnd, this);
 		if (this._docType === 'spreadsheet') {
 			map.on('zoomend', this._onCellCursorShift, this);
 		}
@@ -278,14 +291,17 @@ L.TileLayer = L.GridLayer.extend({
 		for (var key in this._selectionHandles) {
 			this._selectionHandles[key].on('drag dragend', this._onSelectionHandleDrag, this);
 		}
-		this._textArea = map._textArea;
-		this._textArea.focus();
+		this._map._clipboardContainer.focus(true);
 
 		map.setPermission(this.options.permission);
 
 		map.fire('statusindicator', {statusType: 'loleafletloaded'});
 	},
 
+	clearAnnotations: function() {
+		console.debug('Implemented in child  classes');
+	},
+
 	getEvents: function () {
 		var events = {
 			viewreset: this._viewReset,
@@ -437,6 +453,12 @@ L.TileLayer = L.GridLayer.extend({
 		else if (textMsg.startsWith('tile:')) {
 			this._onTileMsg(textMsg, img);
 		}
+		else if (textMsg.startsWith('windowpaint:')) {
+			this._onDialogPaintMsg(textMsg, img);
+		}
+		else if (textMsg.startsWith('window:')) {
+			this._onDialogMsg(textMsg);
+		}
 		else if (textMsg.startsWith('unocommandresult:')) {
 			this._onUnoCommandResultMsg(textMsg);
 		}
@@ -467,19 +489,22 @@ L.TileLayer = L.GridLayer.extend({
 		else if (textMsg.startsWith('editor:')) {
 			this._updateEditor(textMsg);
 		}
+		else if (textMsg.startsWith('validitylistbutton:')) {
+			this._onValidityListButtonMsg(textMsg);
+		}
 	},
 
 	toggleTileDebugMode: function() {
 		this._invalidateClientVisibleArea();
 		this._debug = !this._debug;
 		if (!this._debug) {
-			map.removeLayer(this._debugInfo);
-			map.removeLayer(this._debugInfo2);
+			this._map.removeLayer(this._debugInfo);
+			this._map.removeLayer(this._debugInfo2);
 			$('.leaflet-control-layers-expanded').css('display', 'none');
 		} else {
 			if (this._debugInfo) {
-				map.addLayer(this._debugInfo);
-				map.addLayer(this._debugInfo2);
+				this._map.addLayer(this._debugInfo);
+				this._map.addLayer(this._debugInfo2);
 				$('.leaflet-control-layers-expanded').css('display', 'block');
 			}
 			this._debugInit();
@@ -540,8 +565,8 @@ L.TileLayer = L.GridLayer.extend({
 		parser.href = this._map.options.server;
 
 		var wopiSrc = '';
-		if (map.options.wopiSrc != '') {
-			wopiSrc = '?WOPISrc=' + map.options.wopiSrc;
+		if (this._map.options.wopiSrc != '') {
+			wopiSrc = '?WOPISrc=' + this._map.options.wopiSrc;
 		}
 		var url = this._map.options.webserver + '/' + this._map.options.urlPrefix + '/' +
 		    encodeURIComponent(this._map.options.doc) + '/' + command.jail + '/' + command.dir + '/' + command.name + wopiSrc;
@@ -643,9 +668,17 @@ L.TileLayer = L.GridLayer.extend({
 		if (!this._prevCellCursor) {
 			this._prevCellCursor = L.LatLngBounds.createDefault();
 		}
+		if (!this._cellCursorXY) {
+			this._cellCursorXY = new L.Point(-1, -1);
+		}
+		if (!this._prevCellCursorXY) {
+			this._prevCellCursorXY = new L.Point(-1, -1);
+		}
+
 		if (textMsg.match('EMPTY')) {
 			this._cellCursorTwips = new L.Bounds(new L.Point(0, 0), new L.Point(0, 0));
 			this._cellCursor = L.LatLngBounds.createDefault();
+			this._cellCursorXY = new L.Point(-1, -1);
 		}
 		else {
 			var strTwips = textMsg.match(/\d+/g);
@@ -656,6 +689,7 @@ L.TileLayer = L.GridLayer.extend({
 			this._cellCursor = new L.LatLngBounds(
 							this._twipsToLatLng(topLeftTwips, this._map.getZoom()),
 							this._twipsToLatLng(bottomRightTwips, this._map.getZoom()));
+			this._cellCursorXY = new L.Point(parseInt(strTwips[4]), parseInt(strTwips[5]));
 		}
 
 		var horizontalDirection = 0;
@@ -720,7 +754,10 @@ L.TileLayer = L.GridLayer.extend({
 
 	_onInvalidateCursorMsg: function (textMsg) {
 		var docLayer = this._map._docLayer;
-		var strTwips = textMsg.match(/\d+/g);
+		textMsg = textMsg.substring('invalidatecursor:'.length + 1);
+		var obj = JSON.parse(textMsg);
+		var modifierViewId = parseInt(obj.viewId);
+		var strTwips = obj.rectangle.match(/\d+/g);
 		var topLeftTwips = new L.Point(parseInt(strTwips[0]), parseInt(strTwips[1]));
 		var offset = new L.Point(parseInt(strTwips[2]), parseInt(strTwips[3]));
 		var bottomRightTwips = topLeftTwips.add(offset);
@@ -733,7 +770,7 @@ L.TileLayer = L.GridLayer.extend({
 			this._map.fire('setFollowOff');
 		}
 		this._map.lastActionByUser = false;
-		this._onUpdateCursor();
+		this._onUpdateCursor(this._viewId === modifierViewId);
 	},
 
 	_updateEditor: function(textMsg) {
@@ -1052,6 +1089,7 @@ L.TileLayer = L.GridLayer.extend({
 	},
 
 	_onUnoCommandResultMsg: function (textMsg) {
+		// console.log('_onUnoCommandResultMsg: "' + textMsg + '"');
 		textMsg = textMsg.substring(18);
 		var obj = JSON.parse(textMsg);
 		var commandName = obj.commandName;
@@ -1065,6 +1103,14 @@ L.TileLayer = L.GridLayer.extend({
 		this._map.hideBusy();
 		this._map.fire('commandresult', {commandName: commandName, success: success, result: obj.result});
 
+		if (this._map.CallPythonScriptSource != null) {
+			this._map.CallPythonScriptSource.postMessage(JSON.stringify({'MessageId': 'CallPythonScript-Result',
+										     'SendTime': Date.now(),
+										     'Values': obj
+										    }),
+								     '*');
+			this._map.CallPythonScriptSource = null;
+		}
 	},
 
 	_onRulerUpdate: function (textMsg) {
@@ -1216,6 +1262,26 @@ L.TileLayer = L.GridLayer.extend({
 		}
 	},
 
+	_onDialogPaintMsg: function(textMsg, img) {
+		var command = this._map._socket.parseServerCmd(textMsg);
+
+		this._map.fire('windowpaint', {
+			id: command.id,
+			img: img,
+			width: command.width,
+			height: command.height,
+			rectangle: command.rectangle
+		});
+	},
+
+	_onDialogMsg: function(textMsg) {
+		textMsg = textMsg.substring('window: '.length);
+		var dialogMsg = JSON.parse(textMsg);
+		// e.type refers to signal type
+		dialogMsg.winType = dialogMsg.type;
+		this._map.fire('window', dialogMsg);
+	},
+
 	_onTileMsg: function (textMsg, img) {
 		var command = this._map._socket.parseServerCmd(textMsg);
 		var coords = this._twipsToCoords(command);
@@ -1265,6 +1331,90 @@ L.TileLayer = L.GridLayer.extend({
 				docType: this._docType
 			});
 		}
+		else if (tile && typeof(img) == 'object') {
+			// 'Uint8Array' delta
+			var canvas = document.createElement('canvas');
+			canvas.width = 256;
+			canvas.height = 256;
+			var ctx = canvas.getContext('2d');
+
+			var oldImg = new Image();
+			oldImg.src = tile.el.src;
+			ctx.drawImage(oldImg, 0, 0);
+
+			// FIXME; can we operate directly on the image ?
+			var imgData = ctx.getImageData(0, 0, canvas.width, canvas.height);
+			var oldData = new Uint8ClampedArray(imgData.data);
+
+			var delta = img;
+			var pixSize = canvas.width * canvas.height * 4;
+			var offset = 0;
+
+			console.log('Applying a delta of length ' + delta.length + ' pix size: ' + pixSize + '\nhex: ' + hex2string(delta));
+
+			// Green-tinge the old-Data ...
+//			for (var i = 0; i < pixSize; ++i)
+//			{
+//				oldData[i*4 + 1] = 128;
+//			}
+
+			// wipe to grey.
+//			for (var i = 0; i < pixSize * 4; ++i)
+//			{
+//				imgData.data[i] = 128;
+//			}
+
+			// Apply delta.
+			for (var i = 1; i < delta.length;)
+			{
+				switch (delta[i])
+				{
+				case 99: // 'c': // copy row
+					var count = delta[i+1];
+					var srcRow = delta[i+2];
+					var destRow = delta[i+3];
+					i+= 4;
+					console.log('copy ' + count + ' row(s) ' + srcRow + ' to ' + destRow);
+					for (var cnt = 0; cnt < count; ++cnt)
+					{
+						var src = (srcRow + cnt) * canvas.width * 4;
+						var dest = (destRow + cnt) * canvas.width * 4;
+						for (var j = 0; j < canvas.width * 4; ++j)
+						{
+							imgData.data[dest + j] = oldData[src + j];
+						}
+					}
+					break;
+				case 100: // 'd': // new run
+					destRow = delta[i+1];
+					var destCol = delta[i+2];
+					var span = delta[i+3];
+					offset = destRow * canvas.width * 4 + destCol * 4;
+					i += 4;
+					console.log('apply new span of size ' + span + ' at pos ' + destCol + ', ' + destRow + ' into delta at byte: ' + offset);
+					span *= 4;
+					imgData.data[offset + 1] = 256; // debug - greener start
+					while (span-- > 0) {
+						imgData.data[offset++] = delta[i++];
+					}
+					imgData.data[offset - 2] = 256; // debug - blue terminator
+					break;
+				default:
+					console.log('ERROR: Unknown code ' + delta[i] +
+						    ' at offset ' + i);
+					i = delta.length;
+					break;
+				}
+			}
+
+			ctx.putImageData(imgData, 0, 0);
+
+			tile.oldWireId = tile.wireId;
+			tile.wireId = command.wireId;
+			tile.el.src = canvas.toDataURL('image/png');
+
+			console.log('set new image');
+		}
 		else if (tile) {
 			if (command.wireId != undefined) {
 				tile.oldWireId = command.wireId;
@@ -1335,20 +1485,6 @@ L.TileLayer = L.GridLayer.extend({
 		}
 	},
 
-	_postKeyboardEvents: function(type, charcodes, keycodes) {
-		// Both are arrays
-		if (typeof(charcodes.length) !== 'number' && typeof(keycodes.length) !== 'number')
-			return;
-
-		// both have same length
-		if (charcodes.length !== keycodes.length)
-			return;
-
-		for (var i = 0; i < charcodes.length; i++) {
-			this._postKeyboardEvent(type, charcodes[i], keycodes[i]);
-		}
-	},
-
 	_postKeyboardEvent: function(type, charcode, keycode) {
 		if (this._docType === 'spreadsheet' && this._prevCellCursor && type === 'input') {
 			if (keycode === 1030) { // PgUp
@@ -1383,6 +1519,11 @@ L.TileLayer = L.GridLayer.extend({
 				' char=' + charcode + ' key=' + keycode);
 	},
 
+	// if winId=0, then event is posted on the document
+	_postCompositionEvent: function(winId, type, text) {
+		this._map._socket.sendMessage('textinput id=' + winId + ' type=' + type + ' text=' + encodeURIComponent(text));
+	},
+
 	_postSelectGraphicEvent: function(type, x, y) {
 		this._map._socket.sendMessage('selectgraphic type=' + type +
 				' x=' + x + ' y=' + y);
@@ -1401,12 +1542,24 @@ L.TileLayer = L.GridLayer.extend({
 		return bounds.getSouthWest().equals(new L.LatLng(0, 0)) && bounds.getNorthEast().equals(new L.LatLng(0, 0));
 	},
 
+	_onZoomStart: function () {
+		this._oldCursorOverlayVisibility = this._isCursorOverlayVisible;
+		this._isCursorOverlayVisible = false;
+		this._onUpdateCursor();
+	},
+
+
+	_onZoomEnd: function () {
+		this._isCursorOverlayVisible = this._oldCursorOverlayVisibility;
+		this._onUpdateCursor();
+	},
+
 	// Update cursor layer (blinking cursor).
-	_onUpdateCursor: function (e) {
+	_onUpdateCursor: function (scroll) {
 		var cursorPos = this._visibleCursor.getNorthWest();
 		var docLayer = this._map._docLayer;
 
-		if (!e && !this._map.getBounds().contains(this._visibleCursor) && this._isCursorVisible) {
+		if ((scroll !== false) && !this._map.getBounds().contains(this._visibleCursor) && this._isCursorVisible) {
 			var center = this._map.project(cursorPos);
 			center = center.subtract(this._map.getSize().divideBy(2));
 			center.x = Math.round(center.x < 0 ? 0 : center.x);
@@ -1431,7 +1584,7 @@ L.TileLayer = L.GridLayer.extend({
 
 	// enable or disable blinking cursor and  the cursor overlay depending on
 	// the state of the document (if the falgs are set)
-	_updateCursorAndOverlay: function (update) {
+	_updateCursorAndOverlay: function (/*update*/) {
 		if (this._map._permission === 'edit'
 		&& this._isCursorVisible
 		&& this._isCursorOverlayVisible
@@ -1449,6 +1602,7 @@ L.TileLayer = L.GridLayer.extend({
 				this._cursorMarker.setLatLng(cursorPos, pixBounds.getSize().multiplyBy(this._map.getZoomScale(this._map.getZoom())));
 			}
 			this._map.addLayer(this._cursorMarker);
+			this._map._clipboardContainer.setLatLng(this._visibleCursor.getNorthWest());
 		}
 		else if (this._cursorMarker) {
 			this._map.removeLayer(this._cursorMarker);
@@ -1632,7 +1786,7 @@ L.TileLayer = L.GridLayer.extend({
 		}
 		if (e.type === 'dragend') {
 			e.target.isDragged = false;
-			this._textArea.focus();
+			this._map._clipboardContainer.focus(true);
 			this._map.fire('scrollvelocity', {vx: 0, vy: 0});
 		}
 
@@ -1681,7 +1835,7 @@ L.TileLayer = L.GridLayer.extend({
 	_onUpdateCellCursor: function (horizontalDirection, verticalDirection, onPgUpDn) {
 		if (this._cellCursor && !this._isEmptyRectangle(this._cellCursor)) {
 			var mapBounds = this._map.getBounds();
-			if (!mapBounds.contains(this._cellCursor)) {
+			if (!mapBounds.contains(this._cellCursor) && !this._cellCursorXY.equals(this._prevCellCursorXY)) {
 				var scrollX = 0, scrollY = 0;
 				if (onPgUpDn) {
 					var mapHalfHeight = (mapBounds.getNorth() - mapBounds.getSouth()) / 2;
@@ -1719,6 +1873,7 @@ L.TileLayer = L.GridLayer.extend({
 					center.y = Math.round(center.y < 0 ? 0 : center.y);
 					this._map.fire('scrollto', {x: center.x, y: center.y});
 				}
+				this._prevCellCursorXY = this._cellCursorXY;
 			}
 
 			if (onPgUpDn) {
@@ -1728,6 +1883,7 @@ L.TileLayer = L.GridLayer.extend({
 
 			if (this._cellCursorMarker) {
 				this._map.removeLayer(this._cellCursorMarker);
+				this._map.removeLayer(this._dropDownButton);
 			}
 			this._cellCursorMarker = L.rectangle(this._cellCursor, {
 				pointerEvents: 'none',
@@ -1739,10 +1895,64 @@ L.TileLayer = L.GridLayer.extend({
 				return;
 			}
 			this._map.addLayer(this._cellCursorMarker);
+
+			this._addDropDownMarker();
 		}
 		else if (this._cellCursorMarker) {
 			this._map.removeLayer(this._cellCursorMarker);
 		}
+		this._removeDropDownMarker();
+	},
+
+	_onValidityListButtonMsg: function(textMsg) {
+		var strXY = textMsg.match(/\d+/g);
+		var validatedCell = new L.Point(parseInt(strXY[0]), parseInt(strXY[1]));
+		var show = parseInt(strXY[2]) === 1;
+		if (show) {
+			if (this._validatedCellXY && !this._validatedCellXY.equals(validatedCell)) {
+				this._validatedCellXY = null;
+				this._removeDropDownMarker();
+			}
+			this._validatedCellXY = validatedCell;
+			this._addDropDownMarker();
+		}
+		else if (this._validatedCellXY && this._validatedCellXY.equals(validatedCell)) {
+			this._validatedCellXY = null;
+			this._removeDropDownMarker();
+		}
+	},
+
+	_addDropDownMarker: function () {
+		if (this._validatedCellXY && this._cellCursorXY && this._validatedCellXY.equals(this._cellCursorXY)) {
+			var pos = this._cellCursor.getNorthEast();
+			var cellCursorHeightPx = this._twipsToPixels(this._cellCursorTwips.getSize()).y;
+			var dropDownMarker = this._getDropDownMarker(cellCursorHeightPx);
+			dropDownMarker.setLatLng(pos);
+			this._map.addLayer(dropDownMarker);
+		}
+	},
+
+	_removeDropDownMarker: function () {
+		if (!this._validatedCellXY && this._dropDownButton)
+			this._map.removeLayer(this._dropDownButton);
+	},
+
+	_getDropDownMarker: function (height) {
+		if (height) {
+			var maxHeight = 27; // it matches the max height of the same control in core
+			var topMargin = 0;
+			if (height > maxHeight) {
+				topMargin = height - maxHeight;
+				height = maxHeight;
+			}
+			var icon =  L.divIcon({
+				className: 'spreadsheet-drop-down-marker',
+				iconSize: [undefined, height],
+				iconAnchor: [0, -topMargin]
+			});
+			this._dropDownButton.setIcon(icon);
+		}
+		return this._dropDownButton;
 	},
 
 	// Update text selection handlers.
@@ -1815,7 +2025,10 @@ L.TileLayer = L.GridLayer.extend({
 	_onCopy: function (e) {
 		e = e.originalEvent;
 		e.preventDefault();
-		if (this._selectionTextContent) {
+		if (this._map._clipboardContainer.getValue() !== '') {
+			L.Compatibility.clipboardSet(e, this._map._clipboardContainer.getValue());
+			this._map._clipboardContainer.setValue('');
+		} else if (this._selectionTextContent) {
 			L.Compatibility.clipboardSet(e, this._selectionTextContent);
 
 			// remember the copied text, for rich copy/paste inside a document
@@ -1828,7 +2041,10 @@ L.TileLayer = L.GridLayer.extend({
 	_onCut: function (e) {
 		e = e.originalEvent;
 		e.preventDefault();
-		if (this._selectionTextContent) {
+		if (this._map._clipboardContainer.getValue() !== '') {
+			L.Compatibility.clipboardSet(e, this._map._clipboardContainer.getValue());
+			this._map._clipboardContainer.setValue('');
+		} else if (this._selectionTextContent) {
 			L.Compatibility.clipboardSet(e, this._selectionTextContent);
 
 			// remember the copied text, for rich copy/paste inside a document
@@ -1841,16 +2057,12 @@ L.TileLayer = L.GridLayer.extend({
 	_onPaste: function (e) {
 		e = e.originalEvent;
 		e.preventDefault();
-		var pasteString = L.Compatibility.clipboardGet(e);
-		if (pasteString === 'false' || !pasteString || pasteString === this._selectionTextHash) {
-			// If there is nothing to paste in clipboard, no harm in
-			// issuing a .uno:Paste in case there is something internally copied in the document
-			// or if the content of the clipboard did not change, we surely must do a rich paste
-			// instead of a normal paste
-			this._map._socket.sendMessage('uno .uno:Paste');
+
+		if (e.clipboardData) { // Standard
+			this._dataTransferToDocument(e.clipboardData, /* preferInternal = */ true);
 		}
-		else {
-			this._map._socket.sendMessage('paste mimetype=text/plain;charset=utf-8\n' + pasteString);
+		else if (window.clipboardData) { // IE 11
+			this._dataTransferToDocument(window.clipboardData, /* preferInternal = */ true);
 		}
 	},
 
@@ -1873,39 +2085,59 @@ L.TileLayer = L.GridLayer.extend({
 		e = e.originalEvent;
 		e.preventDefault();
 
-		// handle content
-		var types = e.dataTransfer.types;
-		var hasHTML = false;
-		for (var t = 0; !hasHTML && t < types.length; t++) {
-			if (types[t] === 'text/html') {
-				hasHTML = true;
+		this._dataTransferToDocument(e.dataTransfer, /* preferInternal = */ false);
+	},
+
+	_dataTransferToDocument: function (dataTransfer, preferInternal) {
+		// for the paste, we might prefer the internal LOK's copy/paste
+		if (preferInternal === true) {
+			var pasteString = dataTransfer.getData('text/plain');
+			if (!pasteString) {
+				pasteString = dataTransfer.getData('Text'); // IE 11
+			}
+
+			if (pasteString && pasteString === this._selectionTextHash) {
+				this._map._socket.sendMessage('uno .uno:Paste');
+				return;
 			}
 		}
 
-		var handled = false;
-		for (t = 0; !handled && t < types.length; t++) {
-			var type = types[t];
-			if (type === 'text/html') {
-				this._map._socket.sendMessage('paste mimetype=text/html\n' + e.dataTransfer.getData(type));
-				handled = true;
-			}
-			else if (type === 'text/plain' && !hasHTML) {
-				this._map._socket.sendMessage('paste mimetype=text/plain;charset=utf-8\n' + e.dataTransfer.getData(type));
-				handled = true;
-			}
-			else if (type === 'Files') {
-				var files = e.dataTransfer.files;
-				for (var i = 0; i < files.length; ++i) {
-					var file = files[i];
+		var types = dataTransfer.types;
+
+		// first try to transfer images
+		// TODO if we have both Files and a normal mimetype, should we handle
+		// both, or prefer one or the other?
+		for (var t = 0; t < types.length; ++t) {
+			if (types[t] === 'Files') {
+				var files = dataTransfer.files;
+				for (var f = 0; f < files.length; ++f) {
+					var file = files[f];
 					if (file.type.match(/image.*/)) {
 						var reader = new FileReader();
 						reader.onload = this._onFileLoadFunc(file);
 						reader.readAsArrayBuffer(file);
-						handled = true;
 					}
 				}
 			}
 		}
+
+		// now try various mime types
+		var mimeTypes = [
+			['text/rtf', 'text/rtf'],
+			['text/html', 'text/html'],
+			['text/plain', 'text/plain;charset=utf-8'],
+			['Text', 'text/plain;charset=utf-8']
+		];
+
+		for (var i = 0; i < mimeTypes.length; ++i) {
+			for (t = 0; t < types.length; ++t) {
+				if (mimeTypes[i][0] === types[t]) {
+					var blob = new Blob(['paste mimetype=' + mimeTypes[i][1] + '\n', dataTransfer.getData(types[t])]);
+					this._map._socket.sendMessage(blob);
+					return;
+				}
+			}
+		}
 	},
 
 	_onFileLoadFunc: function(file) {
@@ -2083,36 +2315,36 @@ L.TileLayer = L.GridLayer.extend({
 			this._debugData = {};
 			this._debugDataNames = ['tileCombine', 'fromKeyInputToInvalidate', 'ping', 'loadCount'];
 			for (var i = 0; i < this._debugDataNames.length; i++) {
-				this._debugData[this._debugDataNames[i]] = L.control.attribution({prefix: '', position: 'bottomleft'}).addTo(map);
+				this._debugData[this._debugDataNames[i]] = L.control.attribution({prefix: '', position: 'bottomleft'}).addTo(this._map);
 			}
 			this._debugInfo = new L.LayerGroup();
 			this._debugInfo2 = new L.LayerGroup();
 			this._debugAlwaysActive = new L.LayerGroup();
 			this._debugTyper = new L.LayerGroup();
-			map.addLayer(this._debugInfo);
-			map.addLayer(this._debugInfo2);
+			this._map.addLayer(this._debugInfo);
+			this._map.addLayer(this._debugInfo2);
 			var overlayMaps = {
 				'Tile overlays': this._debugInfo,
 				'Screen overlays': this._debugInfo2,
 				'Always active': this._debugAlwaysActive,
 				'Typing': this._debugTyper
 			};
-			L.control.layers({}, overlayMaps, {collapsed: false}).addTo(map);
+			L.control.layers({}, overlayMaps, {collapsed: false}).addTo(this._map);
 
 			this._map.on('layeradd', function(e) {
 				if (e.layer === this._debugAlwaysActive) {
-					map._debugAlwaysActive = true;
+					this._map._debugAlwaysActive = true;
 				} else if (e.layer === this._debugTyper) {
 					this._debugTypeTimeout();
 				} else if (e.layer === this._debugInfo2) {
 					for (var i = 0; i < this._debugDataNames.length; i++) {
-						this._debugData[this._debugDataNames[i]].addTo(map);
+						this._debugData[this._debugDataNames[i]].addTo(this._map);
 					}
 				}
 			}, this);
-			map.on('layerremove', function(e) {
+			this._map.on('layerremove', function(e) {
 				if (e.layer === this._debugAlwaysActive) {
-					map._debugAlwaysActive = false;
+					this._map._debugAlwaysActive = false;
 				} else if (e.layer === this._debugTyper) {
 					clearTimeout(this._debugTypeTimeoutId);
 				} else if (e.layer === this._debugInfo2) {
@@ -2205,15 +2437,15 @@ L.TileLayer = L.GridLayer.extend({
 					rect.setStyle({fillOpacity: opac - 0.04});
 				}
 			}
-			for (var key in this._debugTiles) {
-				var rect = this._debugTiles[key];
+			for (key in this._debugTiles) {
+				rect = this._debugTiles[key];
 				var col = rect.options.fillColor;
-				var opac = rect.options.fillOpacity;
+				opac = rect.options.fillOpacity;
 				if (col === 'blue' && opac >= 0.04 && rect.date + 1000 < +new Date()) {
 					rect.setStyle({fillOpacity: opac - 0.04});
 				}
 			}
-			this._debugTimeoutId = setTimeout(function () { map._docLayer._debugTimeout(); }, 50);
+			this._debugTimeoutId = setTimeout(function () { this._map._docLayer._debugTimeout(); }, 50);
 		}
 	},
 
@@ -2226,7 +2458,7 @@ L.TileLayer = L.GridLayer.extend({
 			this._postKeyboardEvent('input', this._debugLorem.charCodeAt(this._debugLoremPos % this._debugLorem.length), 0);
 		}
 		this._debugLoremPos++;
-		this._debugTypeTimeoutId = setTimeout(function () { map._docLayer._debugTypeTimeout(); }, 50);
+		this._debugTypeTimeoutId = setTimeout(function () { this._map._docLayer._debugTypeTimeout(); }, 50);
 	}
 
 });
