@@ -1,9 +1,9 @@
-/* -*- js-indent-level: 8 -*- */
+/* -*- js-indent-level: 8; fill-column: 100 -*- */
 /*
  * L.Map is the central class of the API - it is used to create a map.
  */
 
-/* global vex $ */
+/* global vex $ _ */
 L.Map = L.Evented.extend({
 
 	options: {
@@ -19,7 +19,8 @@ L.Map = L.Evented.extend({
 		tileWidthTwips: 3840,
 		tileHeightTwips: 3840,
 		urlPrefix: 'lool',
-		wopiSrc: ''
+		wopiSrc: '',
+		cursorURL: 'images/cursors'
 	},
 
 	lastActiveTime: Date.now(),
@@ -52,7 +53,7 @@ L.Map = L.Evented.extend({
 			this.setView(L.latLng(options.center), options.zoom, {reset: true});
 		}
 
-		L.Cursor.imagePath = options.cursorURL || L.Cursor.getCursorURL('cursors');
+		L.Cursor.imagePath = options.cursorURL;
 
 		if (options.webserver === undefined) {
 			var protocol = window.location.protocol === 'file:' ? 'https:' : window.location.protocol;
@@ -86,6 +87,9 @@ L.Map = L.Evented.extend({
 		this._socket = L.socket(this);
 		this._progressBar = L.progressOverlay(this.getCenter(), L.point(150, 25));
 
+		this._clipboardContainer = L.clipboardContainer();
+		this.addLayer(this._clipboardContainer);
+
 		// Inhibit the context menu - the browser thinks that the document
 		// is just a bunch of images, hence the context menu is useless (tdf#94599)
 		this.on('contextmenu', function() {});
@@ -119,7 +123,8 @@ L.Map = L.Evented.extend({
 				this._fireInitComplete('doclayerinit');
 			}
 			if (this._docLayer._docType == 'text') {
-				L.control.ruler({position:'topleft'}).addTo(this);
+				var interactiveRuler = this._permission === 'edit' ? true : false;
+				L.control.ruler({position:'topleft', interactive:interactiveRuler}).addTo(this);
 			}
 		});
 		this.on('updatetoolbarcommandvalues', function(e) {
@@ -165,6 +170,9 @@ L.Map = L.Evented.extend({
 				} else {
 					this._activate();
 				}
+			} else if (this._docLayer) {
+				// remove the comments and changes
+				this._docLayer.clearAnnotations();
 			}
 		}, this);
 	},
@@ -343,6 +351,11 @@ L.Map = L.Evented.extend({
 		return this.on('moveend', this._panInsideMaxBounds);
 	},
 
+	setDocBounds: function (bounds) {
+		bounds = L.latLngBounds(bounds);
+		this.options.docBounds = bounds;
+	},
+
 	panInsideBounds: function (bounds, options) {
 		var center = this.getCenter(),
 		    newCenter = this._limitCenter(center, this._zoom, bounds);
@@ -414,6 +427,13 @@ L.Map = L.Evented.extend({
 			handler.enable();
 		}
 
+		// Check for the special proof-of-concept case where no WOPI is involved but we
+		// still run loleaflet in an iframe of its own and thus need to receive the
+		// postMessage things.
+		if (name === 'wopi' && this.options['notWopiButIframe']) {
+			handler.addHooks();
+		}
+
 		return this;
 	},
 
@@ -500,6 +520,16 @@ L.Map = L.Evented.extend({
 			this.options.maxZoom;
 	},
 
+	getLayerMaxBounds: function () {
+		return L.bounds(this.latLngToLayerPoint(this.options.maxBounds.getNorthWest()),
+			this.latLngToLayerPoint(this.options.maxBounds.getSouthEast()));
+	},
+
+	getLayerDocBounds: function () {
+		return L.bounds(this.latLngToLayerPoint(this.options.docBounds.getNorthWest()),
+			this.latLngToLayerPoint(this.options.docBounds.getSouthEast()));
+	},
+
 	getBoundsZoom: function (bounds, inside, padding) { // (LatLngBounds[, Boolean, Point]) -> Number
 		bounds = L.latLngBounds(bounds);
 
@@ -643,9 +673,9 @@ L.Map = L.Evented.extend({
 
 	focus: function () {
 		console.debug('focus:');
-		if (this._docLayer && document.activeElement !== this._docLayer._textArea) {
+		if (this._docLayer) {
 			console.debug('focus: focussing');
-			this._docLayer._textArea.focus();
+			this._clipboardContainer.focus(true);
 		}
 	},
 
@@ -673,20 +703,19 @@ L.Map = L.Evented.extend({
 			throw new Error('Map container is already initialized.');
 		}
 
-		var textAreaContainer = L.DomUtil.create('div', 'clipboard-container', container.parentElement);
-		this._textArea = L.DomUtil.create('input', 'clipboard', textAreaContainer);
-		this._textArea.setAttribute('type', 'text');
-		this._textArea.setAttribute('autocorrect', 'off');
-		this._textArea.setAttribute('autocapitalize', 'off');
-		this._textArea.setAttribute('autocomplete', 'off');
-		this._textArea.setAttribute('spellcheck', 'false');
 		this._resizeDetector = L.DomUtil.create('iframe', 'resize-detector', container);
 		this._fileDownloader = L.DomUtil.create('iframe', '', container);
 		L.DomUtil.setStyle(this._fileDownloader, 'display', 'none');
 
+		L.DomEvent.addListener(container, 'scroll', this._onScroll, this);
 		container._leaflet = true;
 	},
 
+	_onScroll: function() {
+		this._container.scrollTop = 0;
+		this._container.scrollLeft = 0;
+	},
+
 	_initLayout: function () {
 		var container = this._container;
 
@@ -803,8 +832,7 @@ L.Map = L.Evented.extend({
 
 		L.DomEvent[onOff](this._container, 'click dblclick mousedown mouseup ' +
 			'mouseover mouseout mousemove contextmenu dragover drop ' +
-			'keydown keypress keyup trplclick qdrplclick', this._handleDOMEvent, this);
-		L.DomEvent[onOff](this._textArea, 'copy cut paste keydown keypress keyup compositionstart compositionupdate compositionend textInput', this._handleDOMEvent, this);
+			'trplclick qdrplclick', this._handleDOMEvent, this);
 
 		if (this.options.trackResize && this._resizeDetector.contentWindow) {
 			L.DomEvent[onOff](this._resizeDetector.contentWindow, 'resize', this._onResize, this);
@@ -822,7 +850,7 @@ L.Map = L.Evented.extend({
 
 	_activate: function () {
 		if (this._serverRecycling || this._documentIdle) {
-			return;
+			return false;
 		}
 
 		console.debug('_activate:');
@@ -834,11 +862,17 @@ L.Map = L.Evented.extend({
 				console.debug('sending useractive');
 				this._socket.sendMessage('useractive');
 				this._active = true;
-				if (this._doclayer) {
+				if (this._docLayer) {
 					this._docLayer._onMessage('invalidatetiles: EMPTY', null);
 				}
+
 				if (vex.dialogID > 0) {
 					var id = vex.dialogID;
+
+					var options = vex.getVexByID(id).data().vex;
+					if (!options.overlayClosesOnClick || !options.escapeButtonCloses)
+						return false;
+
 					vex.dialogID = -1;
 					this._startInactiveTimer();
 					this.focus();
@@ -890,7 +924,7 @@ L.Map = L.Evented.extend({
 			vex: options
 		});
 
-		options.$vex.bind('click.vex', function(e) {
+		options.$vex.bind('click.vex', function() {
 			console.debug('_dim: click.vex function');
 			return map._activate();
 		});
@@ -1046,8 +1080,8 @@ L.Map = L.Evented.extend({
 		// Calling from some other place with no real 'click' event doesn't work
 		if (type === 'click') {
 			if (this._permission === 'edit') {
-				this._textArea.blur();
-				this._textArea.focus();
+				this._clipboardContainer.focus(false);
+				this._clipboardContainer.focus(true);
 			}
 
 			// unselect if anything is selected already
