@@ -181,17 +181,16 @@ L.Map.Keyboard = L.Handler.extend({
 		this._map.off('compositionstart compositionupdate compositionend textInput', this._onKeyDown, this);
 	},
 
-	_handleOnKeyDown: function (keyCode, modifier) {
-		if (modifier & this.keyModifier.shift) {
+	_ignoreKeyEvent: function(e) {
+		var shift = e.originalEvent.shiftKey ? this.keyModifier.shift : 0;
+		if (shift && (e.originalEvent.keyCode === 45 || e.originalEvent.keyCode === 46)) {
 			// don't handle shift+insert, shift+delete
 			// These are converted to 'cut', 'paste' events which are
 			// automatically handled by us, so avoid double-handling
-			if (keyCode === 45 || keyCode === 46) {
-				return false;
-			}
+			return true;
 		}
 
-		return this.handleOnKeyDownKeys[keyCode];
+		return false;
 	},
 
 	_setPanOffset: function (pan) {
@@ -238,11 +237,22 @@ L.Map.Keyboard = L.Handler.extend({
 		return this.keymap[keyCode] || keyCode;
 	},
 
-	_onKeyDown: function (e) {
+	_onKeyDown: function (e, keyEventFn, compEventFn, inputEle) {
 		if (this._map.slideShow && this._map.slideShow.fullscreen) {
 			return;
 		}
 		var docLayer = this._map._docLayer;
+		if (!keyEventFn) {
+			// default is to post keyboard events on the document
+			keyEventFn = L.bind(docLayer._postKeyboardEvent, docLayer);
+		}
+		if (!compEventFn) {
+			// document has winid=0
+			compEventFn = L.bind(docLayer._postCompositionEvent, docLayer, 0 /* winid */);
+		}
+		if (!inputEle) {
+			inputEle = this._map._clipboardContainer._textArea;
+		}
 		this.modifier = 0;
 		var shift = e.originalEvent.shiftKey ? this.keyModifier.shift : 0;
 		var ctrl = e.originalEvent.ctrlKey ? this.keyModifier.ctrl : 0;
@@ -287,24 +297,27 @@ L.Map.Keyboard = L.Handler.extend({
 
 		if (e.type === 'compositionstart' || e.type === 'compositionupdate') {
 			this._isComposing = true; // we are starting composing with IME
+			var txt = '';
+			for (var i = 0; i < e.originalEvent.data.length; i++) {
+				txt += e.originalEvent.data[i];
+			}
+			if (txt) {
+				compEventFn('input', txt);
+			}
 		}
 
 		if (e.type === 'compositionend') {
 			this._isComposing = false; // stop of composing with IME
 			// get the composited char codes
-			var compCharCodes = [];
-			for (var i = 0; i < e.originalEvent.data.length; i++) {
-				compCharCodes.push(e.originalEvent.data[i].charCodeAt());
-			}
 			// clear the input now - best to do this ASAP so the input
 			// is clear for the next word
-			this._map._textArea.value = '';
+			inputEle.value = '';
 		}
 
 		if (!this._isComposing && e.type === 'keyup') {
 			// not compositing and keyup, clear the input so it is ready
 			// for next word (or char only)
-			this._map._textArea.value = '';
+			inputEle.value = '';
 		}
 
 		var unoKeyCode = this._toUNOKeyCode(keyCode);
@@ -312,7 +325,7 @@ L.Map.Keyboard = L.Handler.extend({
 		if (this.modifier) {
 			unoKeyCode |= this.modifier;
 			if (e.type !== 'keyup' && (this.modifier !== shift || (keyCode === 32 && !docLayer._isCursorVisible))) {
-				docLayer._postKeyboardEvent('input', charCode, unoKeyCode);
+				keyEventFn('input', charCode, unoKeyCode);
 				e.originalEvent.preventDefault();
 				return;
 			}
@@ -321,16 +334,19 @@ L.Map.Keyboard = L.Handler.extend({
 		if (this._map._permission === 'edit') {
 			docLayer._resetPreFetching();
 
-			if (e.type === 'keydown') {
+			if (this._ignoreKeyEvent(e)) {
+				// key ignored
+			}
+			else if (e.type === 'keydown') {
 				this._keyHandled = false;
 				this._bufferedTextInputEvent = null;
 
-				if (this._handleOnKeyDown(keyCode, this.modifier) && charCode === 0) {
-					docLayer._postKeyboardEvent('input', charCode, unoKeyCode);
+				if (this.handleOnKeyDownKeys[keyCode] && charCode === 0) {
+					keyEventFn('input', charCode, unoKeyCode);
 				}
 			}
 			else if ((e.type === 'keypress' || e.type === 'compositionend') &&
-			         (!this._handleOnKeyDown(keyCode, this.modifier) || charCode !== 0)) {
+			         (!this.handleOnKeyDownKeys[keyCode] || charCode !== 0)) {
 				if (charCode === keyCode && charCode !== 13) {
 					// Chrome sets keyCode = charCode for printable keys
 					// while LO requires it to be 0
@@ -343,9 +359,9 @@ L.Map.Keyboard = L.Handler.extend({
 				}
 				if (e.type === 'compositionend') {
 					// Set all keycodes to zero
-					docLayer._postKeyboardEvents('input', compCharCodes, Array.apply(null, Array(compCharCodes.length)).map(Number.prototype.valueOf, 0));
+					compEventFn('end', '');
 				} else {
-					docLayer._postKeyboardEvent('input', charCode, unoKeyCode);
+					keyEventFn('input', charCode, unoKeyCode);
 				}
 
 				this._keyHandled = true;
@@ -374,13 +390,11 @@ L.Map.Keyboard = L.Handler.extend({
 				if (!this._keyHandled && this._bufferedTextInputEvent) {
 					var textInputData = this._bufferedTextInputEvent.originalEvent.data;
 					charCode = e.originalEvent.keyCode;
-					var compCharCodes = [];
-					for (var i = 0; i < textInputData.length; i++) {
-						compCharCodes.push(textInputData[i].charCodeAt());
+					for (var idx = 0; idx < textInputData.length; idx++) {
+						keyEventFn('input', textInputData[idx].charCodeAt(), 0);
 					}
-					docLayer._postKeyboardEvents('input', compCharCodes, Array.apply(null, Array(compCharCodes.length)).map(Number.prototype.valueOf, 0));
 				}
-				docLayer._postKeyboardEvent('up', charCode, unoKeyCode);
+				keyEventFn('up', charCode, unoKeyCode);
 
 				this._keyHandled = true;
 				this._bufferedTextInputEvent = null;
@@ -406,7 +420,7 @@ L.Map.Keyboard = L.Handler.extend({
 			else if (key in this._panKeys && e.originalEvent.shiftKey &&
 					docLayer._selections.getLayers().length !== 0) {
 				// if there is a selection and the user wants to modify it
-				docLayer._postKeyboardEvent('input', charCode, unoKeyCode);
+				keyEventFn('input', charCode, unoKeyCode);
 			}
 			else if (key in this._zoomKeys) {
 				map.setZoom(map.getZoom() + (e.shiftKey ? 3 : 1) * this._zoomKeys[key]);
@@ -417,6 +431,10 @@ L.Map.Keyboard = L.Handler.extend({
 	},
 
 	_handleCtrlCommand: function (e) {
+		// Control
+		if (e.originalEvent.keyCode == 17)
+			return true;
+
 		if (e.type !== 'keydown' && e.originalEvent.key !== 'c' && e.originalEvent.key !== 'v' && e.originalEvent.key !== 'x' &&
 			/* Safari */ e.originalEvent.keyCode !== 99 && e.originalEvent.keyCode !== 118 && e.originalEvent.keyCode !== 120) {
 			e.originalEvent.preventDefault();
@@ -431,7 +449,7 @@ L.Map.Keyboard = L.Handler.extend({
 		}
 
 		if (e.originalEvent.ctrlKey && e.originalEvent.shiftKey && e.originalEvent.key === '?') {
-			map.showLOKeyboardHelp();
+			this._map.showLOKeyboardHelp();
 			e.originalEvent.preventDefault();
 			return true;
 		}
@@ -504,9 +522,9 @@ L.Map.Keyboard = L.Handler.extend({
 		case 91: // Left Cmd (Safari)
 		case 93: // Right Cmd (Safari)
 			// we prepare for a copy or cut event
-			this._map._docLayer._textArea.value = 'dummy text';
-			this._map._docLayer._textArea.focus();
-			this._map._docLayer._textArea.select();
+			this._map._clipboardContainer.setValue(window.getSelection().toString());
+			this._map._clipboardContainer.focus(true);
+			this._map._clipboardContainer.select();
 			return true;
 		case 80: // p
 			this._map.print();
