@@ -40,10 +40,13 @@
 
 namespace Poco
 {
+    class MemoryInputStream;
     namespace Net
     {
+        class HTTPRequest;
         class HTTPResponse;
     }
+    class URI;
 }
 
 class Socket;
@@ -95,9 +98,12 @@ public:
     static const int DefaultSendBufferSize = 16 * 1024;
     static const int MaximumSendBufferSize = 128 * 1024;
     static std::atomic<bool> InhibitThreadChecks;
+    std::string _clientAddress;
 
-    Socket() :
-        _fd(socket(AF_INET, SOCK_STREAM | SOCK_NONBLOCK, 0)),
+    enum Type { IPv4, IPv6, All };
+
+    Socket(Type type) :
+        _fd(createSocket(type)),
         _sendBufferSize(DefaultSendBufferSize),
         _owner(std::this_thread::get_id())
     {
@@ -112,6 +118,9 @@ public:
         close(_fd);
     }
 
+    /// Create socket of the given type.
+    static int createSocket(Type type);
+
     /// Returns the OS native socket fd.
     int getFD() const { return _fd; }
 
@@ -237,8 +246,8 @@ public:
     {
         if (id != _owner)
         {
-            LOG_DBG("#" << _fd << " Thread affinity set to 0x" << std::hex <<
-                    id << " (was 0x" << _owner << ")." << std::dec);
+            LOG_DBG("#" << _fd << " Thread affinity set to " << Log::to_string(id) <<
+                    " (was " << Log::to_string(_owner) << ").");
             _owner = id;
         }
     }
@@ -256,9 +265,9 @@ public:
         // 0 owner means detached and can be invoked by any thread.
         const bool sameThread = (_owner == std::thread::id(0) || std::this_thread::get_id() == _owner);
         if (!sameThread)
-            LOG_ERR("#" << _fd << " Invoked from foreign thread. Expected: 0x" << std::hex <<
-                    _owner << " but called from 0x" << std::this_thread::get_id() << " (" <<
-                    std::dec << Util::getThreadId() << ").");
+            LOG_ERR("#" << _fd << " Invoked from foreign thread. Expected: " <<
+                    Log::to_string(_owner) << " but called from " <<
+                    std::this_thread::get_id() << " (" << Util::getThreadId() << ").");
 
         assert(sameThread);
     }
@@ -278,8 +287,7 @@ protected:
         setNoDelay();
         _sendBufferSize = DefaultSendBufferSize;
         _owner = std::this_thread::get_id();
-        LOG_DBG("#" << _fd << " Thread affinity set to 0x" << std::hex <<
-                _owner << "." << std::dec);
+        LOG_DBG("#" << _fd << " Thread affinity set to " << Log::to_string(_owner) << ".");
 
 #if ENABLE_DEBUG
         if (std::getenv("LOOL_ZERO_BUFFER_SIZE"))
@@ -300,6 +308,41 @@ private:
     std::thread::id _owner;
 };
 
+class StreamSocket;
+
+/// Interface that handles the actual incoming message.
+class SocketHandlerInterface
+{
+public:
+    virtual ~SocketHandlerInterface() {}
+    /// Called when the socket is newly created to
+    /// set the socket associated with this ResponseClient.
+    /// Will be called exactly once.
+    virtual void onConnect(const std::shared_ptr<StreamSocket>& socket) = 0;
+
+    /// Called after successful socket reads.
+    virtual void handleIncomingMessage(SocketDisposition &disposition) = 0;
+
+    /// Prepare our poll record; adjust @timeoutMaxMs downwards
+    /// for timeouts, based on current time @now.
+    /// @returns POLLIN and POLLOUT if output is expected.
+    virtual int getPollEvents(std::chrono::steady_clock::time_point now,
+                              int &timeoutMaxMs) = 0;
+
+    /// Do we need to handle a timeout ?
+    virtual void checkTimeout(std::chrono::steady_clock::time_point /* now */) {}
+
+    /// Do some of the queued writing.
+    virtual void performWrites() = 0;
+
+    /// Called when the is disconnected and will be destroyed.
+    /// Will be called exactly once.
+    virtual void onDisconnect() {}
+
+    /// Append pretty printed internal state to a line
+    virtual void dumpState(std::ostream& os) { os << "\n"; }
+};
+
 /// Handles non-blocking socket event polling.
 /// Only polls on N-Sockets and invokes callback and
 /// doesn't manage buffers or client data.
@@ -376,9 +419,9 @@ public:
         // 0 owner means detached and can be invoked by any thread.
         const bool sameThread = (!isAlive() || _owner == std::thread::id(0) || std::this_thread::get_id() == _owner);
         if (!sameThread)
-            LOG_ERR("Incorrect thread affinity for " << _name << ". Expected: 0x" << std::hex <<
-                    _owner << " (" << std::dec << Util::getThreadId() << ") but called from 0x" <<
-                    std::hex << std::this_thread::get_id() << std::dec << ", stop: " << _stop);
+            LOG_ERR("Incorrect thread affinity for " << _name << ". Expected: " <<
+                    Log::to_string(_owner) << " (" << Util::getThreadId() <<
+                    ") but called from " << std::this_thread::get_id() << ", stop: " << _stop);
 
         assert(_stop || sameThread);
     }
@@ -428,11 +471,11 @@ public:
                 std::swap(_newCallbacks, invoke);
             }
 
-            for (size_t i = 0; i < invoke.size(); ++i)
+            for (const auto& callback : invoke)
             {
                 try
                 {
-                    invoke[i]();
+                    callback();
                 }
                 catch (const std::exception& exc)
                 {
@@ -496,7 +539,7 @@ public:
         } while (rc == -1 && errno == EINTR);
 
         if (rc == -1 && errno != EAGAIN && errno != EWOULDBLOCK)
-            LOG_SYS("wakeup socket #" << fd << " is closd at wakeup?");
+            LOG_SYS("wakeup socket #" << fd << " is closed at wakeup?");
     }
 
     /// Wakeup the main polling loop in another thread
@@ -527,6 +570,10 @@ public:
         }
     }
 
+    /// Inserts a new websocket to be polled.
+    /// NOTE: The DNS lookup is synchronous.
+    void insertNewWebSocketSync(const Poco::URI &uri, const std::shared_ptr<SocketHandlerInterface>& websocketHandler);
+
     typedef std::function<void()> CallbackFn;
 
     /// Add a callback to be invoked in the polling thread
@@ -610,8 +657,8 @@ private:
             LOG_INF("Starting polling thread [" << _name << "].");
 
             _owner = std::this_thread::get_id();
-            LOG_DBG("Thread affinity of " << _name << " set to 0x" <<
-                    std::hex << _owner << "." << std::dec);
+            LOG_DBG("Thread affinity of " << _name << " set to " <<
+                    Log::to_string(_owner) << ".");
 
             // Invoke the virtual implementation.
             pollingThread();
@@ -653,53 +700,20 @@ protected:
     std::thread::id _owner;
 };
 
-class StreamSocket;
-
-/// Interface that handles the actual incoming message.
-class SocketHandlerInterface
-{
-public:
-    virtual ~SocketHandlerInterface() {}
-    /// Called when the socket is newly created to
-    /// set the socket associated with this ResponseClient.
-    /// Will be called exactly once.
-    virtual void onConnect(const std::shared_ptr<StreamSocket>& socket) = 0;
-
-    /// Called after successful socket reads.
-    virtual void handleIncomingMessage(SocketDisposition &disposition) = 0;
-
-    /// Prepare our poll record; adjust @timeoutMaxMs downwards
-    /// for timeouts, based on current time @now.
-    /// @returns POLLIN and POLLOUT if output is expected.
-    virtual int getPollEvents(std::chrono::steady_clock::time_point now,
-                              int &timeoutMaxMs) = 0;
-
-    /// Do we need to handle a timeout ?
-    virtual void checkTimeout(std::chrono::steady_clock::time_point /* now */) {}
-
-    /// Do some of the queued writing.
-    virtual void performWrites() = 0;
-
-    /// Called when the is disconnected and will be destroyed.
-    /// Will be called exactly once.
-    virtual void onDisconnect() {}
-
-    /// Append pretty printed internal state to a line
-    virtual void dumpState(std::ostream& os) { os << "\n"; }
-};
-
 /// A plain, non-blocking, data streaming socket.
 class StreamSocket : public Socket, public std::enable_shared_from_this<StreamSocket>
 {
 public:
-    /// Create a StreamSocket from native FD and take ownership of handler instance.
-    StreamSocket(const int fd, std::shared_ptr<SocketHandlerInterface> socketHandler) :
+    /// Create a StreamSocket from native FD.
+    StreamSocket(const int fd, bool /* isClient */,
+                 std::shared_ptr<SocketHandlerInterface> socketHandler) :
         Socket(fd),
         _socketHandler(std::move(socketHandler)),
-        _closed(false),
-        _shutdownSignalled(false),
         _bytesSent(0),
-        _bytesRecvd(0)
+        _bytesRecvd(0),
+        _wsState(WSState::HTTP),
+        _closed(false),
+        _shutdownSignalled(false)
     {
         LOG_DBG("StreamSocket ctor #" << fd);
 
@@ -725,6 +739,10 @@ public:
         }
     }
 
+    bool isClosed() const { return _closed; }
+    bool isWebSocket() const { return _wsState == WSState::WS; }
+    void setWebSocket() { _wsState = WSState::WS; }
+
     /// Just trigger the async shutdown.
     virtual void shutdown() override
     {
@@ -815,20 +833,39 @@ public:
     /// but we can't have a shared_ptr in the ctor.
     template <typename TSocket>
     static
-    std::shared_ptr<TSocket> create(const int fd, std::shared_ptr<SocketHandlerInterface> handler)
+    std::shared_ptr<TSocket> create(const int fd, bool isClient, std::shared_ptr<SocketHandlerInterface> handler)
     {
         SocketHandlerInterface* pHandler = handler.get();
-        auto socket = std::make_shared<TSocket>(fd, std::move(handler));
+        auto socket = std::make_shared<TSocket>(fd, isClient, std::move(handler));
         pHandler->onConnect(socket);
         return socket;
     }
 
+    /// Remove the first @count bytes from input buffer
+    void eraseFirstInputBytes(size_t count)
+    {
+        _inBuffer.erase(_inBuffer.begin(), _inBuffer.begin() + count);
+    }
+
+    /// Detects if we have an HTTP header in the provided message and
+    /// populates a request for that.
+    bool parseHeader(const char *clientLoggingName,
+                     Poco::MemoryInputStream &message,
+                     Poco::Net::HTTPRequest &request,
+                     size_t *requestSize = nullptr);
+
+    /// Get input/output statistics on this stream
     void getIOStats(uint64_t &sent, uint64_t &recv)
     {
         sent = _bytesSent;
         recv = _bytesRecvd;
     }
 
+    const std::string clientAddress()
+    {
+        return _clientAddress;
+    }
+
 protected:
 
     /// Called when a polling event is received.
@@ -962,18 +999,20 @@ protected:
     /// Client handling the actual data.
     std::shared_ptr<SocketHandlerInterface> _socketHandler;
 
+    std::vector<char> _inBuffer;
+    std::vector<char> _outBuffer;
+
+    uint64_t _bytesSent;
+    uint64_t _bytesRecvd;
+
+    enum class WSState { HTTP, WS } _wsState;
+
     /// True if we are already closed.
     bool _closed;
 
     /// True when shutdown was requested via shutdown().
     bool _shutdownSignalled;
 
-    std::vector< char > _inBuffer;
-    std::vector< char > _outBuffer;
-
-    uint64_t _bytesSent;
-    uint64_t _bytesRecvd;
-
     // To be able to access _inBuffer and _outBuffer.
     // TODO we probably need accessors to the _inBuffer & _outBuffer
     // instead of this many friends...
@@ -1004,8 +1043,7 @@ namespace HttpHelper
     void sendFile(const std::shared_ptr<StreamSocket>& socket, const std::string& path, const std::string& mediaType,
                   Poco::Net::HTTPResponse& response, bool noCache = false, bool deflate = false,
                   const bool headerOnly = false);
-};
-
+}
 #endif
 
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
