@@ -7,7 +7,7 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
-#include "config.h"
+#include <config.h>
 
 #include <cassert>
 #include <mutex>
@@ -22,21 +22,22 @@
 #include "Admin.hpp"
 #include "AdminModel.hpp"
 #include "Auth.hpp"
-#include "Common.hpp"
+#include <Common.hpp>
 #include "FileServer.hpp"
-#include "IoUtil.hpp"
+#include <IoUtil.hpp>
 #include "LOOLWSD.hpp"
-#include "Log.hpp"
-#include "Protocol.hpp"
+#include <Log.hpp>
+#include <Protocol.hpp>
 #include "Storage.hpp"
 #include "TileCache.hpp"
-#include "Unit.hpp"
-#include "Util.hpp"
+#include <Unit.hpp>
+#include <Util.hpp>
 
-#include "net/Socket.hpp"
-#include "net/WebSocketHandler.hpp"
+#include <net/Socket.hpp>
+#include <net/SslSocket.hpp>
+#include <net/WebSocketHandler.hpp>
 
-#include "common/SigUtil.hpp"
+#include <common/SigUtil.hpp>
 
 using namespace LOOLProtocol;
 
@@ -44,6 +45,9 @@ using Poco::Net::HTTPResponse;
 using Poco::StringTokenizer;
 using Poco::Util::Application;
 
+const int Admin::MinStatsIntervalMs = 50;
+const int Admin::DefStatsIntervalMs = 2500;
+
 /// Process incoming websocket messages
 void AdminSocketHandler::handleMessage(bool /* fin */, WSOpCode /* code */,
                                        std::vector<char> &payload)
@@ -73,7 +77,7 @@ void AdminSocketHandler::handleMessage(bool /* fin */, WSOpCode /* code */,
         std::string jwtToken;
         LOOLProtocol::getTokenString(tokens[1], "jwt", jwtToken);
         const auto& config = Application::instance().config();
-        const auto sslKeyPath = config.getString("ssl.key_file_path", "");
+        const std::string sslKeyPath = config.getString("ssl.key_file_path", "");
 
         LOG_INF("Verifying JWT token: " << jwtToken);
         JWTAuth authAgent(sslKeyPath, "admin", "admin", "admin");
@@ -158,7 +162,7 @@ void AdminSocketHandler::handleMessage(bool /* fin */, WSOpCode /* code */,
     {
         try
         {
-            const auto pid = std::stoi(tokens[1]);
+            const int pid = std::stoi(tokens[1]);
             LOG_INF("Admin request to kill PID: " << pid);
             SigUtil::killChild(pid);
         }
@@ -225,7 +229,7 @@ void AdminSocketHandler::handleMessage(bool /* fin */, WSOpCode /* code */,
                 {
                     _admin->rescheduleMemTimer(settingVal);
                     model.clearMemStats();
-                    model.notify("settings mem_stats_interval=" + std::to_string(settingVal));
+                    model.notify("settings mem_stats_interval=" + std::to_string(_admin->getMemStatsInterval()));
                 }
             }
             else if (settingName == "cpu_stats_size")
@@ -241,7 +245,7 @@ void AdminSocketHandler::handleMessage(bool /* fin */, WSOpCode /* code */,
                 {
                     _admin->rescheduleCpuTimer(settingVal);
                     model.clearCpuStats();
-                    model.notify("settings cpu_stats_interval=" + std::to_string(settingVal));
+                    model.notify("settings cpu_stats_interval=" + std::to_string(_admin->getCpuStatsInterval()));
                 }
             }
             else if (LOOLProtocol::matchPrefix("limit_", settingName))
@@ -265,6 +269,7 @@ void AdminSocketHandler::handleMessage(bool /* fin */, WSOpCode /* code */,
     }
 }
 
+/// Connection from remote admin socket
 AdminSocketHandler::AdminSocketHandler(Admin* adminManager,
                                        const std::weak_ptr<StreamSocket>& socket,
                                        const Poco::Net::HTTPRequest& request)
@@ -275,6 +280,15 @@ AdminSocketHandler::AdminSocketHandler(Admin* adminManager,
 {
 }
 
+/// Client connection to remote amdin socket
+AdminSocketHandler::AdminSocketHandler(Admin* adminManager)
+    : WebSocketHandler(true),
+      _admin(adminManager),
+      _sessionId(0),
+      _isAuthenticated(true)
+{
+}
+
 void AdminSocketHandler::sendTextFrame(const std::string& message)
 {
     UnitWSD::get().onAdminQueryMessage(message);
@@ -291,10 +305,16 @@ bool AdminSocketHandler::handleInitialRequest(
     const std::weak_ptr<StreamSocket> &socketWeak,
     const Poco::Net::HTTPRequest& request)
 {
-    auto socket = socketWeak.lock();
+    if (!LOOLWSD::AdminEnabled)
+    {
+        LOG_ERR("Request for disabled admin console");
+        return false;
+    }
+
+    std::shared_ptr<StreamSocket> socket = socketWeak.lock();
 
     // Different session id pool for admin sessions (?)
-    const auto sessionId = Util::decodeId(LOOLWSD::GenSessionId());
+    const size_t sessionId = Util::decodeId(LOOLWSD::GenSessionId());
 
     const std::string& requestURI = request.getURI();
     StringTokenizer pathTokens(requestURI, "/", StringTokenizer::TOK_IGNORE_EMPTY | StringTokenizer::TOK_TRIM);
@@ -333,24 +353,24 @@ Admin::Admin() :
     _lastJiffies(0),
     _lastSentCount(0),
     _lastRecvCount(0),
-    _memStatsTaskIntervalMs(5000),
-    _cpuStatsTaskIntervalMs(2000),
-    _networkStatsIntervalMs(5000)
+    _cpuStatsTaskIntervalMs(DefStatsIntervalMs),
+    _memStatsTaskIntervalMs(DefStatsIntervalMs * 2),
+    _netStatsTaskIntervalMs(DefStatsIntervalMs * 2)
 {
     LOG_INF("Admin ctor.");
 
-    _totalSysMem = Util::getTotalSystemMemory();
-    LOG_TRC("Total system memory : " << _totalSysMem);
+    _totalSysMemKb = Util::getTotalSystemMemoryKb();
+    LOG_TRC("Total system memory:  " << _totalSysMemKb << " KB.");
 
     const auto memLimit = LOOLWSD::getConfigValue<double>("memproportion", 0.0);
-    _totalAvailMem = _totalSysMem;
+    _totalAvailMemKb = _totalSysMemKb;
     if (memLimit != 0.0)
-        _totalAvailMem = _totalSysMem * memLimit/100.;
+        _totalAvailMemKb = _totalSysMemKb * memLimit / 100.;
 
-    LOG_TRC("Total available memory: " << _totalAvailMem << " (memproportion: " << memLimit << ").");
+    LOG_TRC("Total available memory: " << _totalAvailMemKb << " KB (memproportion: " << memLimit << "%).");
 
-    const auto totalMem = getTotalMemoryUsage();
-    LOG_TRC("Total memory used: " << totalMem);
+    const size_t totalMem = getTotalMemoryUsage();
+    LOG_TRC("Total memory used: " << totalMem << " KB.");
     _model.addMemStats(totalMem);
 }
 
@@ -371,99 +391,112 @@ void Admin::pollingThread()
 
     while (!_stop && !TerminationFlag && !ShutdownRequestFlag)
     {
-        std::chrono::steady_clock::time_point now = std::chrono::steady_clock::now();
+        const std::chrono::steady_clock::time_point now = std::chrono::steady_clock::now();
 
         int cpuWait = _cpuStatsTaskIntervalMs -
             std::chrono::duration_cast<std::chrono::milliseconds>(now - lastCPU).count();
-        if (cpuWait <= 0)
+        if (cpuWait <= MinStatsIntervalMs / 2) // Close enough
         {
-            size_t currentJiffies = getTotalCpuUsage();
-            auto cpuPercent = 100 * 1000 * currentJiffies / (sysconf (_SC_CLK_TCK) * _cpuStatsTaskIntervalMs);
+            const size_t currentJiffies = getTotalCpuUsage();
+            const size_t cpuPercent = 100 * 1000 * currentJiffies / (sysconf (_SC_CLK_TCK) * _cpuStatsTaskIntervalMs);
             _model.addCpuStats(cpuPercent);
 
-            lastCPU = now;
             cpuWait += _cpuStatsTaskIntervalMs;
+            lastCPU = now;
         }
 
         int memWait = _memStatsTaskIntervalMs -
             std::chrono::duration_cast<std::chrono::milliseconds>(now - lastMem).count();
-        if (memWait <= 0)
+        if (memWait <= MinStatsIntervalMs / 2) // Close enough
         {
-            const auto totalMem = getTotalMemoryUsage();
+            const size_t totalMem = getTotalMemoryUsage();
+            _model.addMemStats(totalMem);
+
             if (totalMem != _lastTotalMemory)
             {
-                LOG_TRC("Total memory used: " << totalMem);
+                // If our total memory consumption is above limit, cleanup
+                triggerMemoryCleanup(totalMem);
+
                 _lastTotalMemory = totalMem;
             }
 
-            _model.addMemStats(totalMem);
-
-            lastMem = now;
             memWait += _memStatsTaskIntervalMs;
-
-            // If our total memory consumption is above limit, cleanup
-            triggerMemoryCleanup(totalMem);
+            lastMem = now;
         }
 
-        int netWait = _networkStatsIntervalMs -
+        int netWait = _netStatsTaskIntervalMs -
             std::chrono::duration_cast<std::chrono::milliseconds>(now - lastNet).count();
-        if(netWait <= 0)
+        if (netWait <= MinStatsIntervalMs / 2) // Close enough
         {
-            uint64_t sentCount = _model.getSentBytesTotal();
-            uint64_t recvCount = _model.getRecvBytesTotal();
+            const uint64_t sentCount = _model.getSentBytesTotal();
+            const uint64_t recvCount = _model.getRecvBytesTotal();
 
             _model.addSentStats(sentCount - _lastSentCount);
             _model.addRecvStats(recvCount - _lastRecvCount);
 
-            LOG_TRC("Total Data sent: " << sentCount << ", recv: " << recvCount);
-            _lastRecvCount = recvCount;
-            _lastSentCount = sentCount;
+            if (_lastRecvCount != recvCount || _lastSentCount != sentCount)
+            {
+                LOG_TRC("Total Data sent: " << sentCount << ", recv: " << recvCount);
+                _lastRecvCount = recvCount;
+                _lastSentCount = sentCount;
+            }
 
+            netWait += _netStatsTaskIntervalMs;
             lastNet = now;
-            netWait += _networkStatsIntervalMs;
         }
+
+        // (re)-connect (with sync. DNS - urk) to one monitor at a time
+        if (_pendingConnects.size())
+        {
+            MonitorConnectRecord rec = _pendingConnects[0];
+            if (rec._when < now)
+            {
+                _pendingConnects.erase(_pendingConnects.begin());
+                connectToMonitorSync(rec._uri);
+            }
+        }
+
         // Handle websockets & other work.
-        int timeout = std::min(std::min(cpuWait, memWait), netWait);
-        LOG_TRC("Admin poll for " << timeout << "ms");
+        const int timeout = capAndRoundInterval(std::min(std::min(cpuWait, memWait), netWait));
+        LOG_TRC("Admin poll for " << timeout << "ms.");
         poll(timeout);
     }
 }
 
 void Admin::modificationAlert(const std::string& dockey, Poco::Process::PID pid, bool value){
-    addCallback([this, dockey, pid, value]
-                { _model.modificationAlert(dockey, pid, value); });
+    addCallback([=] { _model.modificationAlert(dockey, pid, value); });
 }
 
 void Admin::addDoc(const std::string& docKey, Poco::Process::PID pid, const std::string& filename,
         const std::string& sessionId, const std::string& userName, const std::string& userId)
 {
-    addCallback([this, docKey, pid, filename, sessionId, userName, userId]
-                 { _model.addDocument(docKey, pid, filename, sessionId, userName, userId); });
+    addCallback([=] { _model.addDocument(docKey, pid, filename, sessionId, userName, userId); });
 }
 
 void Admin::rmDoc(const std::string& docKey, const std::string& sessionId)
 {
-    addCallback([this, docKey, sessionId]
-                 { _model.removeDocument(docKey, sessionId); });
+    addCallback([=] { _model.removeDocument(docKey, sessionId); });
 }
 
 void Admin::rmDoc(const std::string& docKey)
 {
     LOG_INF("Removing complete doc [" << docKey << "] from Admin.");
-    addCallback([this, docKey]{ _model.removeDocument(docKey); });
+    addCallback([=]{ _model.removeDocument(docKey); });
 }
 
 void Admin::rescheduleMemTimer(unsigned interval)
 {
-    _memStatsTaskIntervalMs = interval;
-    LOG_INF("Memory stats interval changed - New interval: " << interval);
+    _memStatsTaskIntervalMs = capAndRoundInterval(interval);
+    LOG_INF("Memory stats interval changed - New interval: " << _memStatsTaskIntervalMs);
+    _netStatsTaskIntervalMs = capAndRoundInterval(interval); // Until we support modifying this.
+    LOG_INF("Network stats interval changed - New interval: " << _netStatsTaskIntervalMs);
     wakeup();
 }
 
 void Admin::rescheduleCpuTimer(unsigned interval)
 {
-    _cpuStatsTaskIntervalMs = interval;
-    LOG_INF("CPU stats interval changed - New interval: " << interval);
+    _cpuStatsTaskIntervalMs = capAndRoundInterval(interval);
+    LOG_INF("CPU stats interval changed - New interval: " << _cpuStatsTaskIntervalMs);
     wakeup();
 }
 
@@ -487,7 +520,7 @@ size_t Admin::getTotalCpuUsage()
     const size_t wsdJ = Util::getCpuUsage(Poco::Process::id());
     const size_t kitsJ = _model.getKitsJiffies();
 
-    if(_lastJiffies == 0)
+    if (_lastJiffies == 0)
     {
         _lastJiffies = forkitJ + wsdJ;
         return 0;
@@ -511,7 +544,7 @@ unsigned Admin::getCpuStatsInterval()
 
 unsigned Admin::getNetStatsInterval()
 {
-    return _networkStatsIntervalMs;
+    return _netStatsTaskIntervalMs;
 }
 
 AdminModel& Admin::getModel()
@@ -521,19 +554,17 @@ AdminModel& Admin::getModel()
 
 void Admin::updateLastActivityTime(const std::string& docKey)
 {
-    addCallback([this, docKey]{ _model.updateLastActivityTime(docKey); });
+    addCallback([=]{ _model.updateLastActivityTime(docKey); });
 }
 
 void Admin::updateMemoryDirty(const std::string& docKey, int dirty)
 {
-    addCallback([this, docKey, dirty]
-                 { _model.updateMemoryDirty(docKey, dirty); });
+    addCallback([=] { _model.updateMemoryDirty(docKey, dirty); });
 }
 
 void Admin::addBytes(const std::string& docKey, uint64_t sent, uint64_t recv)
 {
-    addCallback([this, docKey, sent, recv]
-                 { _model.addBytes(docKey, sent, recv); });
+    addCallback([=] { _model.addBytes(docKey, sent, recv); });
 }
 
 void Admin::notifyForkit()
@@ -550,56 +581,52 @@ void Admin::notifyForkit()
         LOG_INF("Forkit write pipe not set (yet).");
 }
 
-void Admin::triggerMemoryCleanup(size_t totalMem)
+void Admin::triggerMemoryCleanup(const size_t totalMem)
 {
-    LOG_TRC("Total memory we are consuming (in kB): " << totalMem);
     // Trigger mem cleanup when we are consuming too much memory (as configured by sysadmin)
     const auto memLimit = LOOLWSD::getConfigValue<double>("memproportion", 0.0);
-    LOG_TRC("Mem proportion for LOOL configured : " << memLimit);
-    if (memLimit == 0.0 || _totalSysMem == 0)
+    if (memLimit == 0.0 || _totalSysMemKb == 0)
     {
-        LOG_TRC("Not configured to do memory cleanup. Skipping memory cleanup.");
+        LOG_TRC("Total memory consumed: " << totalMem <<
+                " KB. Not configured to do memory cleanup. Skipping memory cleanup.");
         return;
     }
 
-    float memToFreePercentage = 0;
-    if ( (memToFreePercentage = (totalMem/static_cast<double>(_totalSysMem)) - memLimit/100.) > 0.0 )
-    {
-        int memToFree = memToFreePercentage * _totalSysMem;
-        LOG_TRC("Memory to be freed (in kB) : " << memToFree);
-        // prepare document list sorted by most idle times
-        std::list<DocBasicInfo> docList = _model.getDocumentsSortedByIdle();
+    LOG_TRC("Total memory consumed: " << totalMem << " KB. Configured LOOL memory proportion: " <<
+            memLimit << "% (" << static_cast<size_t>(_totalSysMemKb * memLimit / 100.) << " KB).");
 
-        LOG_TRC("Checking saved documents in document list, length: " << docList.size());
-        // Kill the saved documents first
-        std::list<DocBasicInfo>::iterator docIt = docList.begin();
-        while (docIt != docList.end() && memToFree > 0)
+    const double memToFreePercentage = (totalMem / static_cast<double>(_totalSysMemKb)) - memLimit / 100.;
+    int memToFreeKb = static_cast<int>(memToFreePercentage > 0.0 ? memToFreePercentage * _totalSysMemKb : 0);
+    // Don't kill documents to save a KB or two.
+    if (memToFreeKb > 1024)
+    {
+        // prepare document list sorted by most idle times
+        const std::vector<DocBasicInfo> docList = _model.getDocumentsSortedByIdle();
+
+        LOG_TRC("OOM: Memory to free: " << memToFreePercentage << "% (" <<
+                memToFreeKb << " KB) from " << docList.size() << " docs.");
+
+        for (const auto& doc : docList)
         {
-            LOG_TRC("Document: DocKey[" << docIt->_docKey << "], Idletime[" << docIt->_idleTime << "],"
-                    << " Saved: [" << docIt->_saved << "], Mem: [" << docIt->_mem << "].");
-            if (docIt->_saved)
+            LOG_TRC("OOM Document: DocKey: [" << doc.DocKey << "], Idletime: [" << doc.IdleTime << "]," <<
+                    " Saved: [" << doc.Saved << "], Mem: [" << doc.Mem << "].");
+            if (doc.Saved)
             {
-                // Kill and remove from list
-                LOG_DBG("OOM: Killing saved document with DocKey " << docIt->_docKey);
-                LOOLWSD::closeDocument(docIt->_docKey, "oom");
-                memToFree -= docIt->_mem;
-                docIt = docList.erase(docIt);
+                // Kill the saved documents first.
+                LOG_DBG("OOM: Killing saved document with DocKey [" << doc.DocKey << "] with " << doc.Mem << " KB.");
+                LOOLWSD::closeDocument(doc.DocKey, "oom");
+                memToFreeKb -= doc.Mem;
+                if (memToFreeKb <= 1024)
+                    break;
             }
             else
-                ++docIt;
-        }
-
-        // Save unsaved documents
-        docIt = docList.begin();
-        while (docIt != docList.end() && memToFree > 0)
-        {
-            LOG_TRC("Saving document: DocKey[" << docIt->_docKey << "].");
-            LOOLWSD::autoSave(docIt->_docKey);
-            ++docIt;
+            {
+                // Save unsaved documents.
+                LOG_TRC("Saving document: DocKey [" << doc.DocKey << "].");
+                LOOLWSD::autoSave(doc.DocKey);
+            }
         }
     }
-
-    LOG_TRC("OOM: Memory to free percentage : " << memToFreePercentage);
 }
 
 void Admin::dumpState(std::ostream& os)
@@ -608,4 +635,87 @@ void Admin::dumpState(std::ostream& os)
     SocketPoll::dumpState(os);
 }
 
+class MonitorSocketHandler : public AdminSocketHandler
+{
+    bool _connecting;
+    std::string _uri;
+public:
+
+    MonitorSocketHandler(Admin *admin, const std::string &uri) :
+        AdminSocketHandler(admin),
+        _connecting(true),
+        _uri(uri)
+    {
+    }
+    int getPollEvents(std::chrono::steady_clock::time_point now,
+                      int &timeoutMaxMs) override
+    {
+        if (_connecting)
+        {
+            LOG_TRC("Waiting for outbound connection to complete");
+            return POLLOUT;
+        }
+        else
+            return AdminSocketHandler::getPollEvents(now, timeoutMaxMs);
+    }
+
+    void performWrites() override
+    {
+        LOG_TRC("Outbound monitor - connected");
+        _connecting = false;
+        return AdminSocketHandler::performWrites();
+    }
+
+    void onDisconnect() override
+    {
+        LOG_WRN("Monitor " << _uri << " dis-connected, re-trying in 20 seconds");
+        Admin::instance().scheduleMonitorConnect(_uri, std::chrono::steady_clock::now() + std::chrono::seconds(20));
+    }
+};
+
+void Admin::connectToMonitorSync(const std::string &uri)
+{
+    insertNewWebSocketSync(Poco::URI(uri), std::make_shared<MonitorSocketHandler>(this, uri));
+}
+
+void Admin::scheduleMonitorConnect(const std::string &uri, std::chrono::steady_clock::time_point when)
+{
+    assertCorrectThread();
+
+    MonitorConnectRecord todo;
+    todo._when = when;
+    todo._uri = uri;
+    _pendingConnects.push_back(todo);
+}
+
+void Admin::start()
+{
+    bool haveMonitors = false;
+    const auto& config = Application::instance().config();
+
+    for (size_t i = 0; ; ++i)
+    {
+        const std::string path = "monitors.monitor[" + std::to_string(i) + "]";
+        const std::string uri = config.getString(path, "");
+        if (!config.has(path))
+            break;
+        if (!uri.empty())
+        {
+            Poco::URI monitor(uri);
+            if (monitor.getScheme() == "wss" || monitor.getScheme() == "ws")
+            {
+                addCallback([=] { scheduleMonitorConnect(uri, std::chrono::steady_clock::now()); });
+                haveMonitors = true;
+            }
+            else
+                LOG_ERR("Unhandled monitor URI: '" << uri << "' should be \"wss://foo:1234/baa\"");
+        }
+    }
+
+    if (!haveMonitors)
+        LOG_TRC("No monitors configured.");
+
+    startThread();
+}
+
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
